/*****************************************************************************
* xor                                                                        *
* Copyright 2006 Nathan Rosenquist                                           *
*                                                                            *
* https://nathanrosenquist.com/xor/                                          *
*                                                                            *
* Takes source and key input and either XORs or modularly adds them          *
* together. If the key data is completely random and well guarded, this can  *
* be used to implement One Time Pad (OTP) encryption                         *
*                                                                            *
* Two modes are offered, XOR and the classic mode which only supports the    *
* letters A-Z. Run the program with no arguments for instructions.           *
*                                                                            *
* The algorithm itself is provably correct and 100% secure, the weakness     *
* lies in the distribution, randomness, and secrecy of the keys themselves.  *
*                                                                            *
* For more information on One Time Pads as an encryption scheme, see         *
* http://en.wikipedia.org/wiki/One_time_pad                                  *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define BUF_SIZE 4096
#define COLUMN_WIDTH 78
#define GROUP_WIDTH 5

/*****************************************************************************
*                        EXTERNAL FUNCTION PROTOTYPES                        *
*****************************************************************************/

int feof(FILE *stream);
int fflush (FILE *stream);
size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream);

/*****************************************************************************
*                        INTERNAL FUNCTION PROTOTYPES                        *
*****************************************************************************/

void show_usage(void);

/*****************************************************************************
*                                    MAIN                                    *
*****************************************************************************/

int main(int argc, char *argv[]) {
    /****************************************
    *           DECLARE VARIABLES           *
    ****************************************/

    /*
        operational mode:
        
        e = encrypt (classic)
        d = decrypt (classic)
        x = xor (standard)
    */
    char mode = 'x';
    
    /* start on this command line argument when looking for files */
    int argv_file_counter = 1;
    
    /* filenames */
    char *file1 = NULL;
    char *file2 = NULL;
    
    /* buffer characters */
    char s = '\x00';
    char k = '\x00';
    
    size_t s_result = (size_t)NULL;
    size_t k_result = (size_t)NULL;
    
    /* buffer arrays */
    char s_buf[BUF_SIZE];
    char k_buf[BUF_SIZE];
    
    /* file pointers */
    FILE *fp_source = NULL;
    FILE *fp_key = NULL;
    
    /* loop counters */
    unsigned int i = 0;
    int column = 0;
    int grouping = 0;
    
    /****************************************
    *           DEFINE VARIABLES            *
    ****************************************/
    
    s_buf[0] = '\x00';
    k_buf[0] = '\x00';
    
    /****************************************
    *         CORE PROGRAM STRUCTURE        *
    ****************************************/
    
    /* set the operating mode, by checking the first argument for -e, -d, or -x */
    if (NULL != argv[1]) {
        if ('-' == argv[1][0]) {
            mode = argv[1][1];
            
            if ('x' == mode) {
                /* xor mode */
                
            } else if ('e' == mode) {
                /* encrypt classic mode */
                
            } else if ('d' == mode) {
                /* decrypt classic mode */
                
            } else {
                /* unknown mode */
                fprintf(stderr, "xor: invalid mode\n");
                exit(1);
            }
            
            /* if we're still here, we have a mode of some sort, so don't use the command line arg as a filename */
            argv_file_counter++;
        }
    }
    
    /* assign command line options to files */
    file1 = argv[argv_file_counter];
    if (NULL != argv[argv_file_counter+1]) {
        file2 = argv[argv_file_counter+1];
    }
    
    /* print usage information (and exit) if we don't have at least a key file */
    if ((argc < 1) || (NULL == file1)) {
        show_usage();
    }
    
    /* open 'key' file */
    fp_key = fopen(file1, "rb");
    if (NULL == fp_key) {
        fprintf(stderr, "xor: could not open key file\n");
        exit(1);
    }
    
    /* open 'source' file if we have one */
    if (NULL != file2) {
        fp_source = fopen(file2, "rb");
        if (NULL == fp_source) {
            fprintf(stderr, "xor: could not open source file\n");
            
            fclose(fp_key);
            exit(1);
        }
        
    /* otherwise use stdin instead */
    } else {
        fp_source = stdin;
    }
    
    /* classic method (A-Z), encrypt or decrypt */
    /* for simplicity's sake, we're reading one character at a time here, instead of reading buffered blocks from files */
    if (('e' == mode) || ('d' == mode)) {
        
        column = 0;
        while (1) {
            s = '\x00';
            k = '\x00';
            
            /* read the next capitalized alphabetical character from 'source', skipping all others */
            while ((s < 'A') || (s > 'Z')) {
                s = fgetc(fp_source);
                if (feof(fp_source)) {
                    break;
                }
            }
            if (feof(fp_source)) {
                break;
            }
            
            /* read the next capitalized alphabetical character from the 'key' file (first arg) */
            /* by doing this after the 'source' data stream, we'll know right away if we're out of key data */
            while ((k < 'A') || (k > 'Z')) {
                k = fgetc(fp_key);
                if (feof(fp_key)) {
                    break;
                }
            }
            if (feof(fp_key)) {
                /* flush output so the error is the last thing that's seen */
                fflush(stdout);
                
                fprintf(stderr, "xor: ran out of key data!\n");
                
                fclose(fp_key);
                if (NULL != file2) {
                    fclose(fp_source);
                }
                
                exit(1);
            }
            
            /* encrypt */
            if ('e' == mode) {
                /* add the source and key letters together */
                s += k;
                
                /* shift back into ASCII capital letter realm */
                s -= 'A';
                
                /* if adding the two letters still caused it to wrap, bring it back in range (26 letters in the alphabet) */
                if (s > 'Z') {
                    s -= 26;
                }
                
            /* decrypt */
            } else {
                /* subtract the key from the source */
                s -= k;
                
                /* shift back into ASCII capital letter realm */
                s += 'A';
                
                /* if subtracting the two letters still caused it to wrap, bring it back in range (26 letters in the alphabet) */
                if (s < 'A') {
                    s += 26;
                }
            }
            
            /* print enciphered character to stdout */
            putchar(s);
            
            /* keep track of where we are */
            column++;
            grouping++;
            
            /* print a space between each 5 characters */
            if (grouping > (GROUP_WIDTH - 1)) {
                putchar(' ');
                grouping = 0;
                column++;
            }
            
            /* print out in 80 column format */
            if (column >= COLUMN_WIDTH) {
                printf("\n");
                grouping = 0;
                column = 0;
            }
        }
        
        /* if we didn't just print a newline, finish off the message with one */
        if (column > 0) {
            printf("\n");
        }
        
    /* XOR method */
    } else {
        while (!feof(fp_source)) {
            /* read characters from the 'source' data stream... */
            s_result = fread(&s_buf, sizeof(char), BUF_SIZE, fp_source);
            
            /* read characters from the 'key' file (first arg) */
            /* by doing this after the 'source' data stream, we'll know right away if we're out of key data */
            k_result = fread(&k_buf, sizeof(char), BUF_SIZE, fp_key);
            if (k_result < s_result) {
                /* flush output so the error is the last thing that's seen */
                fflush(stdout);
                
                fprintf(stderr, "xor: ran out of key data!\n");
                
                fclose(fp_key);
                if (NULL != file2) {
                    fclose(fp_source);
                }
                
                exit(1);
            }
            
            /* loop through source and key buffers together */
            for (i=0; i<s_result; i++) {
                /* XOR source character against the key and print to stdout */
                putchar(s_buf[(int)i] ^ k_buf[(int)i]);
            }
        }
    }
    
    /* close all the files that we opened */
    fclose(fp_key);
    if (NULL != fp_source) {
        fclose(fp_source);
    }
    
    return 0;
}

/*****************************************************************************
*                                 FUNCTIONS                                  *
*****************************************************************************/

void show_usage(void) {
    fprintf(stderr, "------------------------------------------------------------------------------\n");
    fprintf(stderr, "xor 1.0.0\n");
    fprintf(stderr, "Usage: xor [-edx] keyfile [sourcefile] > outfile\n");
    fprintf(stderr, "------------------------------------------------------------------------------\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "xor implements both symmetrical binary XOR encryption/decryption, and the\n");
    fprintf(stderr, "classic One Time Pad encryption using only the capital letters A-Z.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "You must specify a keyfile to use on the command line, and the source\n");
    fprintf(stderr, "data to be encrypted/decrypted can come from either a file specified\n");
    fprintf(stderr, "as the second argument, or from standard input.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "The default mode of operation is XOR, which allows completely symmetrical\n");
    fprintf(stderr, "binary encryption/decryption both ways.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "The classic One Time Pad using only capital letters is also implemented,\n");
    fprintf(stderr, "and requires the -e or -d flags for encrypt or decrypt mode.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "The keyfile must contain either a series of truly random bits (such as\n");
    fprintf(stderr, "from a hardware random number generator, /dev/random isn't good enough),\n");
    fprintf(stderr, "or a series of random capital letters in the case of classic A-Z encryption.\n");
    fprintf(stderr, "The security of either method is only as good as the random numbers used.\n");
    fprintf(stderr, "Keys should be used only once, and then destroyed by both the sender and\n");
    fprintf(stderr, "recipient.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Some examples:\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "  XOR (binary) encryption:\n");
    fprintf(stderr, "    xor keyfile [sourcefile] > encrypted_file\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "  XOR (binary) decryption:\n");
    fprintf(stderr, "    xor keyfile [sourcefile] > original_file\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "  OTP classic A-Z encryption (key and source data must be all capital letters):\n");
    fprintf(stderr, "    xor -e keyfile [sourcefile] > encrypted_file\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "  OTP classic A-Z encryption (key and source data must be all capital letters):\n");
    fprintf(stderr, "    xor -d keyfile [sourcefile] > original_file\n");
    fprintf(stderr, "\n");
    
    exit(1);
}

